// Copyright (c) 2014, Google Inc.
// SPDX-License-Identifier: ISC

// This header is provided for compatibility with older revisions of BoringSSL.
// TODO(davidben): Remove this header.

#include "crypto.h"
