// Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0 OR ISC
// Mon Jun 29 12:41:07 UTC 2026

pub(super) const CRYPTO_LIBRARY: &[&str] = &[
    "generated-src/linux-ppc64le/crypto/fipsmodule/aesp8-ppc.S",
    "generated-src/linux-ppc64le/crypto/fipsmodule/ghashp8-ppc.S",
    "generated-src/linux-ppc64le/crypto/test/trampoline-ppc.S",
];
